/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */
package org.objectweb.jonas.jmx;

// JMX includes
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.management.MBeanServer;
import javax.management.MBeanServerFactory;

// JOnAS includes
import org.objectweb.jonas.common.Log;
import org.objectweb.jonas.service.AbsServiceImpl;
import org.objectweb.jonas.service.ServiceException;

// Monolog
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;


/**
 * JMX Service implementation
 * This singleton class must exist in each jonas server that is to be administered via JMX
 * @author Guillaume Riviere
 * Contributor(s): Michel Bruno, Adriana Danes
 * Adriana Danes : implement new version of this service allowing to use mx4j.
 *                 highlight configuration properties
 */

public abstract class AbsJmxServiceImpl extends AbsServiceImpl implements JmxService {

    // Trace logger
     static protected Logger logger = null;

    // the name of the JOnAS server that is administerd by the MBeanServer
    protected String jonasName = null;

    // MBeanServer instance
    protected MBeanServer jmxServer = null;

    // Jmx service configuration parameters
    public static final String CLASS = "jonas.service.jmx.class";

    /**
     * Init the Service.
     * Create the MBeanServer.
     * Configuration information is passed through a Context object.
     */
    public void doInit(Context ctx) throws ServiceException {
        logger = Log.getLogger(Log.JONAS_SERVER_PREFIX);

	// Create a MBeanServer with a standard default domain name
	// The default domain name is used as the domain part in the ObjectName of MBeans. 
	// The standard default domain name is defined in ServiceName.DOMAIN (currently, DefaultDomain)
	// jmxServer = MBeanServerFactory.createMBeanServer();

	// Create a MBeanServer with "jonas" default domain name 
	String defaultDomain = JonasObjectName.DOMAIN;
	jmxServer = MBeanServerFactory.createMBeanServer(defaultDomain);
	
	// Gets the configuration
	try {
	    jonasName = (String) ctx.lookup("jonas.name");
	} catch (NamingException ne) {
	    logger.log(BasicLevel.DEBUG, "Cannot initialize JMX service "+ne);
	    throw new ServiceException("Cannot initialize JMX service", ne);
	}
	logger.log(BasicLevel.DEBUG, "JMX Service initialized");
    }

    /**
     * Start the Service
     * Initialization of the service is already done.
     */
    public abstract void doStart() throws ServiceException ;

    /**
     * Stop this service
     */
    public abstract void doStop() throws ServiceException ;

    /**
     * @return The reference of the MBean server
     */
    public MBeanServer getJmxServer() {
	return this.jmxServer;
    }

    /**
     * Remove internal references to the MBeanServer. 
     */
    protected void releaseJmxServer() {
	try {
	    MBeanServerFactory.releaseMBeanServer(jmxServer); 
	} catch(java.lang.IllegalArgumentException ex) {
	    // The mbeanServer is not found.
	    logger.log(BasicLevel.ERROR, "Cannot find the MBeanServer" + ex);
	}
    }
}
