/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.jmx.vjadmin.servlet;

import java.util.Iterator;
import java.util.Set;
import java.util.Vector;
import java.util.Hashtable;
import javax.management.ObjectName;
import javax.management.MalformedObjectNameException;
import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;
import org.apache.velocity.VelocityContext;
import org.apache.velocity.context.Context;
import org.apache.velocity.servlet.VelocityServlet;
import org.objectweb.jonas.common.Log;
import org.objectweb.jonas.jmx.JonasObjectName;
import org.objectweb.jonas.jmx.vjadmin.JonasManagementRepr;
import org.objectweb.jonas.jmx.vjadmin.ManagementException;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;


/**
 * Abstract class inherited by all servlets managing EJBs
 * @author Philippe Durieux, Adriana Danes (JOnAS team)
 * Contributor(s):
 *
 * 03/01/14 Adriana Danes
 *       New version of getDatasourceInfo() method. 
 */
public abstract class EjbServlet extends JadminServlet {

    /**
     * Get info about a bean
     */
    void getBeanInfo(Context ctx, ObjectName mbean) {
	ctx.put("EJBClass", (String) JonasManagementRepr.getAttribute(mbean, "EJBClass"));
	ctx.put("DisplayName", (String) JonasManagementRepr.getAttribute(mbean, "DisplayName"));
	ctx.put("JndiName", (String) JonasManagementRepr.getAttribute(mbean, "JndiName"));
	String home = (String) JonasManagementRepr.getAttribute(mbean, "HomeClass");
	ctx.put("HomeClass", home != null ? home : "none");
	String remote = (String) JonasManagementRepr.getAttribute(mbean, "RemoteClass");
	ctx.put("RemoteClass", remote != null ? remote : "none");
	String localhome = (String) JonasManagementRepr.getAttribute(mbean, "LocalHomeClass");
	ctx.put("LocalHomeClass", localhome != null ? localhome : "none");
	String local = (String) JonasManagementRepr.getAttribute(mbean, "LocalClass");
	ctx.put("LocalClass", local != null ? local : "none");
    }

    /**
     * Get information about DataSources referenced by the bean
     */
    void getDatasourceInfo(Context ctx, ObjectName mbean) {
	String error = "";
	String error_dbmService = "";
	try {
	   // Get data sources
	    Hashtable dataSources = (Hashtable)JonasManagementRepr.getAttribute(mbean, "AllDataSourceName");
	    ctx.put("dataSources", dataSources);
	    ctx.put("nbDatasources", new Integer(dataSources.size()));
	} catch (ManagementException me) {
	    error_dbmService = "DataBase Service Not Available";
	} catch (Exception e) {
	    error = "Error while getting attribute:" + e.getMessage();
	}
	ctx.put("datasource_error", error);
	ctx.put("error_dbmService", error_dbmService);
    }

    /**
     * Get information about JMS Connection Factories referenced by the bean
     */
    void getConnFactoryInfo(Context ctx, ObjectName mbean) {
	String error = "";
	try {
	    // Get JMS Connection Factories
	    Set allJmsConnFactories = (Set)JonasManagementRepr.getAttribute(mbean, "AllJMSConnectionFactoryName");
	    int nbConnFactories = allJmsConnFactories.size();
	    ctx.put("nbConnFactories", new Integer(nbConnFactories));
	    if (nbConnFactories > 0) {
		ctx.put("jmsConnFactories", allJmsConnFactories.toArray());
		// verify JMS service is activated
		boolean activatedJMS = JonasManagementRepr.isRegistered(JonasObjectName.jmsService());
		ctx.put("activatedJMS", new Boolean(activatedJMS));
	    }
	} catch (Exception e) {
	    error = "Error while getting attribute:" + e.getMessage();
	}
	ctx.put("jmsConnFact_error", error);
    }

    /**
     * Get information about JMS Destinations referenced by the bean
     */
    void getDestinationInfo(Context ctx, ObjectName mbean) {
	String error = "";
	try {
	    // Get JMS Destinations
	    Set allDestinations = (Set) JonasManagementRepr.getAttribute(mbean, "AllJMSDestinationName");
	    int nbDest = allDestinations.size();
	    ctx.put("nbDestinations", new Integer(nbDest));
	    if (nbDest > 0) {
		ctx.put("destinations", allDestinations.toArray());
		// verify JMS service is activated
		boolean activatedJMS = JonasManagementRepr.isRegistered(JonasObjectName.jmsService());
		ctx.put("activatedJMS", new Boolean(activatedJMS));
	    }		
	} catch (Exception e) {
	    error = "Error while getting managed attribute:" + e.getMessage();
	}
	ctx.put("jmsDest_error", error);
    }
    /**
     * Get information about Mail Factories referenced by the bean
     */
    void getMailFactoryInfo(Context ctx, ObjectName mbean) {
	String error = "";
	boolean activatedMail = false;
	try {
	    // Get Session Mail Factories
	    Hashtable sessMailFactories = (Hashtable)JonasManagementRepr.getAttribute(mbean, "AllMailFactorySName");
	    ctx.put("sessMailFactories", sessMailFactories);
	    int nbFact = sessMailFactories.size();
	    ctx.put("nbSessMailFactories", new Integer(nbFact));
	    if (nbFact > 0) {
		// verify if Mail Service activated
		activatedMail = JonasManagementRepr.isRegistered(JonasObjectName.mailService());
		ctx.put("activatedMail", new Boolean(activatedMail));
	    }

	    // Get MimePartDS Mail Factories
	    Hashtable mimeMailFactories = (Hashtable)JonasManagementRepr.getAttribute(mbean, "AllMailFactoryMName");
	    ctx.put("mimeMailFactories", mimeMailFactories);
	    nbFact = mimeMailFactories.size();
	    ctx.put("nbMimeMailFactories", new Integer(nbFact));
	    if (nbFact > 0) {
		// verify if Mail Service activated
		activatedMail = JonasManagementRepr.isRegistered(JonasObjectName.mailService());
		ctx.put("activatedMail", new Boolean(activatedMail));
	    }	    
	} catch (Exception e) {
	    error = "Error while getting managed attribute:" + e.getMessage();
	}
	ctx.put("mailFact_error", error);
    }
}
