/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.jtm;

import java.rmi.NoSuchObjectException;
import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.io.OutputStream;
import java.io.FileOutputStream;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.File;
import java.util.Properties;
import javax.management.ObjectName;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.transaction.TransactionManager;
import javax.transaction.UserTransaction;
import org.objectweb.jonas_tm.Current;
import org.objectweb.jonas_tm.TransactionFactory;
import org.objectweb.jonas_tm.TransactionFactoryImpl;
import org.objectweb.jonas.common.Log;
import org.objectweb.jonas.jmx.JmxService;
import org.objectweb.jonas.jmx.JonasObjectName;
import org.objectweb.common.Env;
import org.objectweb.jonas.naming.NamingManager;
import org.objectweb.jonas.service.AbsServiceImpl;
import org.objectweb.jonas.service.ServiceManager;
import org.objectweb.jonas.service.ServiceException;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * Transaction Service implementation.
 * This singleton class must exist in each jonas server.
 * This class manages a unique Current object that implements both
 * TransactionManager and UserTransaction interfaces.
 * @author Philippe Durieux
 * Contributor(s):
 */
public class TransactionServiceImpl extends AbsServiceImpl implements TransactionService, TransactionServiceImplMBean {

    static private Logger logger = null;

    // Transaction Service configuration properties
    public static final String TIMEOUT = "jonas.service.jtm.timeout";
    public static final String REMOTE = "jonas.service.jtm.remote";
    public static final String CLASS = "jonas.service.jtm.class";
    public static final String SERVER = "jonas.name"; // gives the JOnAS server name
    public static final String JONAS = "jonas";// default JOnAS server name

    // TM factory may be local or remote.
    // The TM factory is used by the JTA implementation when transactions
    // are distributed among several JVM (JOnAS Server or applet client)
    private TransactionFactory tm = null;

    // Unique Current object implementing the standard TransactionManager
    // set to null while service is not started.
    private Current current = null;

    // Configuration information used to start the Transaction service
    private int timeout;
    private boolean jtmlocal;
    private InitialContext ictx;

    // Other configuration information for the Transaction service
    private String jtmServiceImplClassName = null;
    private String jonasServName = null;

    // Boolean equal to false until a configuration information gets modified via a MBean method
    private boolean reconfigured = false;
 
    // Keeps the up to date configuration parameters (contains reconfiguration values)
    Properties configProps = new Properties();

    // -------------------------------------------------------------------
    // JOnAS Service Implementation
    // -------------------------------------------------------------------

    /**
     * Init the Service.
     * Configuration information is passed thru a Context object.
     */
    public void doInit(Context ctx) throws ServiceException {

	// get logger for this service
        logger = Log.getLogger(Log.JONAS_SERVER_PREFIX);

	// Gets the configuration
	String remote = "false";
	try {
	    remote = (String) ctx.lookup(REMOTE);
	} catch (NamingException e) {
	    // No problem if there is no value --> default value
	}
	if ((remote!=null) && (!remote.equals(""))) {
	    jtmlocal = !remote.equalsIgnoreCase("true");
	} else {
	    jtmlocal = false;
	}

	String tstr = "60";
	try {
	    tstr = (String) ctx.lookup(TIMEOUT);
	} catch (NamingException e) {
	    // No problem if there is no value --> default value
	}
	if ((tstr!=null) && (!tstr.equals(""))) {
	    timeout = (new Integer(tstr)).intValue();
	} else {
	    timeout = 60;
	}

	try {
	    jtmServiceImplClassName = (String) ctx.lookup(CLASS);
	} catch (NamingException e) {
	    // No problem if there is no value --> default value
	    jtmServiceImplClassName = this.getClass().getName();
	}

	try {
	    jonasServName = (String) ctx.lookup(SERVER);
	} catch (NamingException e) {
	    // Get default value jonas (??)
	    jonasServName = JONAS;
	}

	logger.log(BasicLevel.DEBUG, "TransactionService initialized");
    }

    /**
     * Start the Service
     * Initialization of the service is already done.
     */
    public void doStart() throws ServiceException {

	try {
	    if (Env.isDavid()) {
		// use CosNaming
		ictx = NamingManager.getInstance().getCorbaInitialContext();
	    } else {
		// use default naming
		ictx = NamingManager.getInstance().getInitialContext();
	    }
	} catch (NamingException e) {
	    logger.log(BasicLevel.ERROR, "TransactionService: Cannot get InitialContext:\n"+e);
	    throw new ServiceException("TransactionService: Cannot get InitialContext", e);
	}

	// Gets the Distributed Transaction Manager (JTM)
	if (jtmlocal) {
	    // Creates a JTM locally in this server
	    logger.log(BasicLevel.DEBUG, "working with a colocated Transaction Manager ");

	    // TODO: Recovery of the JTM

	    // Creates the ControlFactory and register it in JNDI.
	    try {
		logger.log(BasicLevel.DEBUG, "Create and register TM factory");
		tm = new TransactionFactoryImpl();
		ictx.rebind("TMFactory", (Object) tm);
	    } catch (RemoteException e) {
		logger.log(BasicLevel.ERROR, "TransactionService: Cannot create TransactionFactory:\n"+e);
		throw new ServiceException("TransactionService: Cannot create TransactionFactory",
					   e);
	    } catch (NamingException e) {
		logger.log(BasicLevel.ERROR, "TransactionService: Cannot rebind TM:\n"+e);
		throw new ServiceException("TransactionService: Cannot rebind TM", e);
	    }
	} else {
	    // JTM is remote, finds it by JNDI
	    logger.log(BasicLevel.DEBUG, "working with a remote Transaction Manager ");
	    final int maxloops = 5;
	    for (int i = 0; i <= maxloops; i++) {
		try {
		    tm = (TransactionFactory) ictx.lookup("TMFactory");
		    break;
		} catch (NamingException e) {
		    if (i < maxloops) {
			logger.log(BasicLevel.WARN, "Cannot get TM factory - retrying...");
			try {
			    Thread.currentThread().sleep(2000*(i+1));
			} catch (InterruptedException e2) {
			    throw new ServiceException("Cannot get TM factory", e2);
			}
		    } else {
			logger.log(BasicLevel.ERROR, "TransactionService: Cannot get TM factory:\n"+e);
			throw new ServiceException("TransactionService: Cannot get TM factory", e);
		    }
		}
	    }
	}

	// Create and init the unique Current object.
	// Current is the local TransactionManager. It must be present in every
	// jonas server and implements the JTA TransactionManager interface.
	//
	// In case of a JTM standalone: 
	// We must create the current only to export it via JNDI for clients
	// that want to get UserTransaction.
	current = new Current(tm, Env.isDavid());
	setTimeout(timeout);
	
	try {
	    // Register TransactionService MBean : JmxTransactionService 
	    ((JmxService)ServiceManager.getJmxService()).getJmxServer().registerMBean(this, JonasObjectName.transactionService());
	}  catch (ServiceException se) {
	    // Jmx Service not available, do nothing
	}  catch (Exception e) {
	    logger.log(BasicLevel.ERROR, "TransactionService: Cannot start the Transaction service:\n"+e);
	    throw new ServiceException("TransactionService: Cannot start the Transaction service", e);
	}

	logger.log(BasicLevel.DEBUG, "TransactionService started, default timeout= "+timeout);
    }

    /**
     * Stop the transaction service
     * Not already implementated
     */
    public void doStop() throws ServiceException {
	logger.log(BasicLevel.DEBUG, "Stop of TransactionService not already implemented");
    }

    // -------------------------------------------------------------------
    // TransactionService Implementation
    // -------------------------------------------------------------------

    /**
     * Gets the Current object instance
     */
    public Current getCurrent() {
	return current;
    }

    /**
     * Gets the TransactionManager object instance
     */
    public TransactionManager getTransactionManager() {
	return (TransactionManager) current;
    }

    /**
     * Gets the UserTransaction object instance
     */
    public UserTransaction getUserTransaction() {
	return (UserTransaction) current;
    }

    /**
     * Gets the TransactionFactory object (JTM factory)
     */
    public TransactionFactory getTransactionFactory() {
	return tm;
    }

    /**
     * Sets the default transaction timeout and register Current in JNDI
     */
    public void setTimeout(int t) {

	logger.log(BasicLevel.DEBUG, "entering for "+t);
	current.setDefaultTimeout(t);

	// Register a UserTransactionFactory in JNDI.
	// Only if we are inside the JTM
	if (jtmlocal) {
	    try {
		logger.log(BasicLevel.DEBUG, "Register UserTransactionFactory");
		NamingManager.getInstance().getInitialContext().rebind("java:comp/UserTransaction", current);
	    } catch (NamingException e) {
		logger.log(BasicLevel.ERROR, "Cannot rebind UserTransaction:"+e);
	    }
	}

    }

    /**
     * MBean  method
     * set the default timeout
     * mark the transaction service as reconfigured
     */
    public void setCurrentTimeOut(Integer t) {
	logger.log(BasicLevel.DEBUG, "entering");
	setTimeout(t.intValue());
	// the property named TIMEOT changed into
	String oldValue = (String)configProps.setProperty(TIMEOUT, String.valueOf(t));
	if (oldValue != null) {
	    logger.log(BasicLevel.DEBUG, "Reconfigured timeout from value : " + oldValue + " to value : " + String.valueOf(t));
	} else {
	    String iniValue = String.valueOf(timeout);
	    logger.log(BasicLevel.DEBUG, "Reconfigured timeout from initial value : " + iniValue + " to value : " + String.valueOf(t));
	}
	reconfigured = true;
    }

    /**
     * MBean  method
     * get the default timeout
     */
    public Integer getCurrentTimeOut() {
	logger.log(BasicLevel.DEBUG, "entering");
	return new Integer(current.getDefaultTimeout());
    }

    /**
     * MBean  method
     * @return true if the jtm is local
     */
    public Boolean getJTMLocal() {
	return new Boolean(jtmlocal);
    }

    /**
     * MBean  method	 
     * @return the jtm port number
     */
    public Integer getJTMPortNumber() {
	try {
	    return new Integer(tm.getPortNumber());
	} catch (Exception e) {
	    return null;
	}
    }  

    /**
     * MBean  method	 
     * @return the jtm host name
     */
    public String getJTMHostName() {
	try {
	    return tm.getHostName();
	} catch (Exception e) {
	    return null;
	}
    }

    /**
     *  MBean method:
     * @return all tx counter
     */
    public Integer [] getTxCounters() {
	return current.getTxCounters();
    }

   /**
     * MBean method:
     * reset all tx counter
     */
    public void resetAllTxTotalCounters() {
	current.resetAllTxTotalCounters();
    }

    /**
     * MBean method: saves the updated server configuration.
     * The modified configuration values are saved in the jonas.properties conf. file
     * located in :
     * <li>
     * <ul>the current start-up directory (.)</ul>
     * <ul>the directory $JONAS_ROOT/update/jonasServName, where <i>jonasServName</i>
     *     is the name of the current JOnAS server as given by the <code>jonas.name</code>
     *     configuration property. 
     *     Currently, this location sould not be used because the JOnAS server bootstrap
     *     was not modified to take into account this location for config files.
     * </li>
     */
    public void saveConfig() {
	logger.log(BasicLevel.DEBUG, "entering");
	if (reconfigured) {
	    OutputStream os = null;
	    // Choose the location for saving the configuration file
	    //boolean local = false;
	    boolean local = true;
	    try {
		os = getOutputStream(local);
	    } catch (Exception e) {
		logger.log(BasicLevel.WARN, "Management operation SaveConfig failed - could not get OutputStream (" + e.toString() + ")");
		return;
	    }

	    String header = "Saved Transaction Service configuration properties via Jadmin ";
	    try {		    
		configProps.store(os, header);
		reconfigured = false;
		configProps = new Properties();
	    } catch (IOException ioe) {
		logger.log(BasicLevel.WARN, "Management operation SaveConfig failed - could not store the up-dated configuration properties (" + ioe.toString() + ")");
		// Try to close os
		try {
		    os.close();
		} catch (IOException ce) {
		    logger.log(BasicLevel.WARN, "Could not close OutputStream (" + ce.toString() + ")"); 
		}
	    }
	}
    }

    private OutputStream getOutputStream(boolean local) throws Exception {
	String propFileName = null;
	if (local) {	    
	    // Save configuration in the ./jonas.properties file
	    // Configuration values will be written to the end of the file
	    propFileName = "jonas.properties";
	} else {
	    Properties systEnv = System.getProperties();
	    String installRoot = systEnv.getProperty("install.root");
	    String fileSeparator = systEnv.getProperty("file.separator");
	    if ((installRoot != null) && (fileSeparator != null)) {
		String propFilePath = installRoot.trim()+fileSeparator+"config"+fileSeparator+"update"+fileSeparator+jonasServName;
		File f = new File(propFilePath);
		if (!f.exists()) {
		    f.mkdirs();
		    logger.log(BasicLevel.DEBUG, " --- " + propFilePath + " created --- ");
		}
		propFileName = propFilePath+fileSeparator+"jonas.properties";
	    } else {
		// try to save in local
		propFileName = "jonas.properties";
	    }
	}
        logger.log(BasicLevel.DEBUG, " --- Save to file " + propFileName + " ---");
	return new FileOutputStream(propFileName, true);
    }

}

