/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or 1any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Florent BENOIT & Ludovic BERT
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.mail.factory;

//import java
import java.io.ObjectOutputStream;
import java.io.ByteArrayOutputStream;
import java.io.Serializable;
import java.rmi.RemoteException;
import java.util.Enumeration;
import java.util.Properties;
import java.util.StringTokenizer;
import org.apache.regexp.RE;
import org.apache.regexp.RESyntaxException;

//import javax
import javax.naming.NamingException;
import javax.naming.Reference;
import javax.naming.Referenceable;

//import objectweb.util
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

//import jonas
import org.objectweb.jonas.service.ServiceManager;
import org.objectweb.jonas.mail.MailService;
import org.objectweb.jonas.common.Log;
import org.objectweb.jonas.common.PropDump;

/**
 * This class implements JOnAS mail factory objects. It gets the properties from mail factory properties file
 * and build a properties object for the Session.
 * @author Florent Benoit
 * @author Ludovic Bert
 * Contributor(s):
 * Adriana Danes : Refactor code: rename the management methods (use straightforward
 * names, as they are to be used in a Management Console)
 * Markus Karg : Change the JOnAS mail factory initialisation strategy in order
 * to allow any initialisation parameters for a javax.mail.Session object (not
 * only a known set pf initialisation parameters as before).
 */
public abstract class JMail implements Serializable, Referenceable, JMailMBean {

    /**
     * The logger used in JOnAS
     */
    protected static Logger logger = null;

    /**
     * The name of the factory.
     */
    private String factoryName = null;

    /**
     * The jndi name of this object.
     */
    private String name = null;

    /**
     * Properties for the javax.mail.Session object.
     * Keep only the values used for configuring the javax.mail.Session object.
     */
    protected Properties mailSessionProperties = null;
    
    /**
     * Properties for the authentication.
     */
    protected Properties authenticationProperties = null;

    /**
     * JOnAS-specific property that authenticates the user.
     */
    private final static String PROPERTY_AUTHENTICATION_USERNAME = "mail.authentication.username";
    private final static String PROPERTY_AUTHENTICATION_PASSWORD = "mail.authentication.password";

    /**
     * Constructor of a JMail Object with the given name and properties.
     * @param factoryName the name of the factory.
     * @param name the jndi name.
     * @param mailProperties properties for configuring and manageing this object.
     */
    public JMail(String factoryName, String name, Properties mailProperties) {
	//Get the logger
        if (logger == null) {
            logger = Log.getLogger(Log.JONAS_MAIL_PREFIX);
        }
	this.factoryName = factoryName;
        this.name = name;

	this.mailSessionProperties = mailProperties;

	// Construct authentication properties from well-known JOnAS properties only
	this.authenticationProperties = new Properties();
	String propValue = null;
	propValue = (String)mailSessionProperties.remove(PROPERTY_AUTHENTICATION_USERNAME);
	if (propValue != null)
	    authenticationProperties.setProperty(PROPERTY_AUTHENTICATION_USERNAME, propValue);	
	propValue = (String)mailSessionProperties.remove(PROPERTY_AUTHENTICATION_PASSWORD);
	if (propValue != null)
	    authenticationProperties.setProperty(PROPERTY_AUTHENTICATION_PASSWORD, propValue);

    }


    /**
     * Return the jndi name of this object
     * @return jndi name of this Jmail.
     */
    public String _getName() {
        return name;
    }

    /**
     * Set the jndi name of this object.
     * @param jndi name 
     */
    void _setName(String name) {
	String oldName = this.name;
	this.name = name;
	((MailService)ServiceManager.getMailService()).renameMailFactory(oldName, this);
    }

    /**
     * Return the name of this mail factory
     * @return name of this mail factory.
     */
    public String _getFactoryName() {
        return factoryName;
    }

    /**
     * Return the type of the factory
     * @return the type of the mail factory
     */
    public abstract String _getType();


    /**
     * Retrieves the session properties of this object. 
     * @return the session properties of this object.
     */
    Properties _getMailSessionProperties() {
        return mailSessionProperties;
    }

    /**
     * Set the session properties.
     * @param props the session properties.
     */
    void _setSessionProperties(Properties props) {
        this.mailSessionProperties = props;
    }

    /**
     * Retrieves the authentication properties of this object. 
     * @return the authentication properties of this object.
     */
    Properties _getAuthenticationProperties() {
        return authenticationProperties;
    }

    /**
     * Set the authentication properties.
     * @param props the authentication properties.
     */
    void _setAuthenticationProperties(Properties props) {
        this.authenticationProperties = props;
    }

    /**
     * Retrieves the Reference of the object. 
     * The Reference contains the factory used to create this object 
     * and the optional parameters used to configure the factory.
     * @return the non-null Reference of the object. 
     * @throws NamingException if a naming exception was encountered while 
     * retrieving the reference.
     */
    public abstract Reference getReference() throws NamingException;


    /**
     * Return an array of byte from a given object
     * @param obj the object from which we must extract the bytes.
     * @return the byte[] from an object
     */
    protected byte[] getBytesFromObject(Object obj) {

        if (obj == null) {
            return null;
        }

        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        
        ObjectOutputStream oos = null;
        byte[] bytes = null;
      
        try {
            oos = new ObjectOutputStream(baos);
            oos.writeObject(obj);
            bytes = baos.toByteArray();
        } catch (Exception e) {
            //Can't tranform
            return null;
        } finally {
            try {
                oos.close();
                baos.close();
            } catch (Exception e) {
                //Not a problem
            }
        }
        return bytes;
    }

    // --------------//
    // MBean methods //
    // --------------//

    /**
     * Get the name of the factory.
     * @return the name of the mail factory (the ressource name as defined in jonas.properties)
     */
    public String getFactoryName() {
        return this._getFactoryName();
    }

    /**
     * Get the jndi name of the factory.
     * @return the jndi name of the mail factory. Should be equal to the one
     * provided in the JOnAS specific Deployment Descriptor of the EJBs using the Mail Service.
     */
    public String getName() {
        return this._getName();
    }
    /**
     * Set the jndi name of the factory.
     * @param jndi name of the mail factory. Give the name
     * provided in the JOnAS specific Deployment Descriptor of the EJBs using the Mail Service.
     */
    public void setName(String name) {
	this._setName(name);
    }

    /**
     * Get the type of the factory.
     * @return type of the mail factory.
     * @throws RemoteException if rmi call failed.
     */
    public String getFactoryType() {
        return this._getType();
    }

    /**
     * Get the authentication properties.
     * @return properties of the authentication.
     */
    public Properties getAuthenticationProperties() {
        return this._getAuthenticationProperties();
    }

    /**
     * Set the authentication properties.
     * @param props the authentication properties.
     */
    public void setAuthenticationProperties(Properties props) {
        this._setAuthenticationProperties(props);
	PropDump.print("These are the udated auth properties",
		       this.authenticationProperties, logger, BasicLevel.DEBUG);
	((MailService)ServiceManager.getMailService()).recreateMailFactory(this);
    }

    /**
     * Get the session properties.
     * @return the properties of Session.
     */
    public Properties getSessionProperties() {
        return this._getMailSessionProperties();
    }

    /**
     * Set the session properties.
     * @param props the session properties.
     */
    public void setSessionProperties(Properties props) {	    
        this._setSessionProperties(props);
	PropDump.print("These are the udated session properties",
		       this.mailSessionProperties, logger, BasicLevel.DEBUG);
	((MailService)ServiceManager.getMailService()).recreateMailFactory(this);
    }




}

