/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Florent BENOIT & Ludovic BERT
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.mail.factory;

//import java
import java.rmi.RemoteException;
import java.util.Enumeration;
import java.util.Properties;

//import javax
import javax.naming.BinaryRefAddr;
import javax.naming.NamingException;
import javax.naming.Reference;

//import jonas
import org.objectweb.jonas.service.ServiceManager;
import org.objectweb.jonas.mail.MailService;
import org.objectweb.jonas.common.PropDump;

//import objectweb.util
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * This class provides a way for referencing mail session.
 * @author Florent Benoit
 * @author Ludovic Bert
 */
public class JMailMimePartDS extends JMail implements JMailMimePartDSMBean {

    /**
     * Type of the factory
     */
    private static final String FACTORY_TYPE = "javax.mail.internet.MimePartDataSource";

    /**
     * Properties for the javax.mail.MimeMessage object.
     */
    private Properties messageProperties = null;

    /**
     * JOnAS-specific property for Mime Messages configuration
     */
    protected static final String MIMEMESSAGE_TO = "mail.to";
    protected static final String MIMEMESSAGE_CC = "mail.cc";
    protected static final String MIMEMESSAGE_BCC = "mail.bcc";
    protected static final String MIMEMESSAGE_SUBJECT = "mail.subject";
 
    /**
     * Return the type of the factory
     * @return the type of the mail factory
     */
    public String _getType() {
        return FACTORY_TYPE;
    }
    

    /**
     * Constructor of a JMimePartDS Object
     * @param factoryName the name of the factory. 
     * @param name the name of this object.
     * @param mailProperties properties for configuring this object.
     */
    public JMailMimePartDS(String factoryName, String name, Properties mailProperties) {
        super(factoryName, name, mailProperties);

        PropDump.print("Received props:", mailProperties, logger, BasicLevel.DEBUG);

        //Get the message properties
        messageProperties = new Properties();
	String propValue = null;
	propValue = (String)mailProperties.remove(MIMEMESSAGE_TO);
	if (propValue != null)
	    messageProperties.setProperty(MIMEMESSAGE_TO, propValue);
	propValue = (String)mailProperties.remove(MIMEMESSAGE_CC);
	if (propValue != null)
	    messageProperties.setProperty(MIMEMESSAGE_CC, propValue);
	propValue = (String)mailProperties.remove(MIMEMESSAGE_BCC);
	if (propValue != null)
	    messageProperties.setProperty(MIMEMESSAGE_BCC, propValue);
	propValue = (String)mailProperties.remove(MIMEMESSAGE_SUBJECT);	
	if (propValue != null)
	    messageProperties.setProperty(MIMEMESSAGE_SUBJECT, propValue);

	PropDump.print("Message props:", messageProperties, logger, BasicLevel.DEBUG);
    }


    /**
     * Retrieves the mimeMessage properties of this object. 
     * @return the mimeMessage properties of this object.
     */
    Properties _getMimeMessageProperties() {
        return messageProperties;
    }

    /**
     * Set the mimeMessage properties of this object.
     * @param props the mimeMessage properties
     */
    void _setMimeMessageProperties(Properties props) {
        this.messageProperties = props;
    }

 
    /**
      * Retrieves the Reference of the javax.mail.MimePartDataSource object. 
     * The Reference contains the factory used to create this object 
     * (that is the JMimePartDSFactory) and the optional parameters used to 
     * configure the factory.
     * @return the non-null Reference of the javax.mail.MimePartDataSource 
     * object. 
     * @throws NamingException if a naming exception was encountered while 
     * retrieving the reference.
     */
    public Reference getReference() throws NamingException {

        //Build the reference to the factory JMailSessionFactory
        Reference reference = new Reference(FACTORY_TYPE,
                                            "org.objectweb.jonas.mail.factory.JMailMimePartDSFactory", 
                                            null);
        byte[] bytes = null;

        //Give the session properties
        bytes = getBytesFromObject(mailSessionProperties);
        if (bytes != null) {
            reference.add(new BinaryRefAddr("javaxmailSession.properties", bytes));
        }

        //Give the mime message properties
        bytes = getBytesFromObject(messageProperties);
        if (bytes != null) {
            reference.add(new BinaryRefAddr("javaxInternetMimeMessage.properties", bytes));
        }

        //Give the authentication properties
        bytes = getBytesFromObject(authenticationProperties);
        if (bytes != null) {
            reference.add(new BinaryRefAddr("authentication.properties", bytes));
        }

	return reference;

    }

    //MBean methods

    /**
     * Get the mimePartDatasource  properties.
     * @return the properties of Session.
     */
    public Properties getMimeMessageProperties() {
        return this._getMimeMessageProperties();
    }

    /**
     * Set the mimePartDatasource properties.
     * @param props the authentication properties.
     * @throws RemoteException if rmi call failed.
     */
    public void setMimeMessageProperties(Properties props) {
        this._setMimeMessageProperties(props);
	PropDump.print("These are the udated message props",
		       this.messageProperties, logger, BasicLevel.DEBUG);
	((MailService)ServiceManager.getMailService()).recreateMailFactory(this);
    }

}
