/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * ObjectWeb Connector: an implementation of JCA Sun specification along
 *                      with some extensions of this specification.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Author: E. Hardesty
 * 
 * Based on Pool api in ObjectWeb common
 *
 */

/**
 * Package definition.
 */
package org.objectweb.jonas.resource.pool.api;

/**
 * Import clauses: one for each external Java definition used. Do not use
 * clauses in the form "import package.*".
 */


/**
 * The interface <b>Pool</b> defines an object that pools resources of any
 * kind. Resources must be requested (getResource) and released
 * (releaseResource) on demand. A Pool object can be parameterized along
 * different dimensions. All these dimensions are represented by accessor
 * methods (getters and setters) assigned to each of them:
 * <ul>
 * <li>
 * A <b>Timeout</b> can be assigned to a Pool. It is used when no more
 * resources are available and when the Pool has reached its maximum size.
 * This is the timeout to wait for a free resource until exiting with an
 * exception. It defaults to 0, which means waiting forever until a
 * resource is freed. Its value should be greater or equal to 0 ; negative
 * values are ignored (setTimeOut(-1) => NOP).
 * </li>
 * <li>
 * A <b>MinSize</b> can be assigned to a Pool. Its value should be greater or
 * equal to 0, and smaller or equal to MaxSize. Values that do not match these
 * conditions are ignored (e.g., setMinSize(-1) => NOP).
 * This size means that there is always MinSize PoolResource allocated in this
 * Pool. If PoolResource needs to be allocated when setting this size,
 * "getPoolMatchFactory().createResource(null)" is called. Its default value
 * is 0.
 * </li>
 * <li>
 * A <b>MatchFactory</b> (i.e., a PoolMatchFactory object) must be assigned to
 * a Pool. It defines the way new PoolResource are created and the way a
 * PoolResource of a Pool matches some "hints" properties when requested. It is
 * mandatory for the Pool to be fully functional.
 * </li>
 * <li>
 * A <b>MaxSize</b> can be assigned to a Pool. Its value should be greater or
 * equal to 0, and greater or equal to MinSize. Values that do not match these
 * conditions are ignored (e.g., setMaxSize(-1) => NOP). Its default value
 * is 0, thus it is mandatory to set this value for making the Poll functional.
 * </li>
 * </ul>
 */
public interface Pool {
        /**
         * <b>getMatchFactory</b> retrieves the PoolMatchFactory assigned to this
         * Pool.
         * @return        The PoolMatchFactory currently assigned to this Pool.
         */
        PoolMatchFactory getMatchFactory();
        
        /**
         * <b>getMaxSize</b> retrieves the maximum size assigned to this Pool.
         * @return        The maximum size currently assigned to this Pool.
         */
        int getMaxSize();
        
        /**
         * <b>getMinSize</b> retrieves the minimum size assigned to this Pool.
         * @return        The minimum size currently assigned to this Pool.
         */
        int getMinSize();
        
        /**
         * <b>getResource</b> is used to allocate a Object from the Pool.
         * Some hints are passed in order to specialise the matching or creation
         * of Object.
         * @param hints        Some properties to specialise the matching or the creation
         *                                of Object.
         * @return                The Object allocated from the Pool.
         */
        Object getResource(Object hints) throws Exception;
        
        /**
         * <b>getTimeout</b> retrieves the timeout assigned to this Pool.
         * @return        The timeout currently assigned to this Pool.
         */
        long getTimeout();
        
        /**
         * <b>getSize</b> retrieves the current size of this Pool.
         * @return        The current size of this Pool.
         */
        int getSize();
        
        /**
         * <b>releaseResource</b> releases a Object in order to allow the
         * Pool to recycle this Object.
         * @param resource        The Object to be released.
         * @param destroy         boolean to remove the object from the pool and
         *                         destroy it
         */
        void releaseResource(Object resource, boolean destroy) throws Exception;
        
        /**
         * <b>setInitSize</b> creates initsize resoures to this Pool.
         * @param initsize       The init size to be created.
         */
        void setInitSize(int initsize) throws Exception;

        /**
         * <b>setMatchFactory</b> assigns a PoolMatchFactory to this Pool.
         * @param pmf                The PoolMatchFactory to be assigned.
         */
        void setMatchFactory(PoolMatchFactory pmf);
        
        /**
         * <b>setMaxSize</b> assigns a maximum size to this Pool.
         * @param maxsize        The maximum size to be assigned.
         */
        void setMaxSize(int maxsize) throws Exception;

        /**
         * <b>setMinSize</b> assigns a minimum size to this Pool.
         * @param minsize        The minimum size to be assigned.
         */
        void setMinSize(int minsize) throws Exception;
        
        /**
         * <b>setTimeout</b> assigns a timeout to this Pool.
         * @param crto                The timeout to be assigned.
         */
        void setTimeout(long crto);
}


