/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * ObjectWeb Connector: an implementation of JCA Sun specification along
 *                      with some extensions of this specification.
 * Copyright (C) 2001-2002 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0 
 *
 * Author: E. Hardesty
 * 
 * Based on LArrayPool in ObjectWeb common
 *
 */

package org.objectweb.jonas.resource.pool.lib;

import java.util.HashSet;
import java.util.Iterator;

import javax.resource.spi.ManagedConnection;
import org.objectweb.jonas.resource.pool.api.Pool;
import org.objectweb.jonas.resource.pool.api.PoolMatchFactory;

/**
 * The class <b>HArrayPool</b> implements a Pool as a HashSet of ManagedConnections,
 * managing free/active resources.
 */
public class HArrayPool implements Pool {

    private int minSize = 0;
    private int maxSize = 0;
    private int initSize = -1;
    private long timeout = 0;
    private int curSize = 0;
    private int freeSize = 0;
    private PoolMatchFactory matchFactory = null;
    private HashSet freeList = null;
    private HashSet activeList = null;
        
    public HArrayPool() {
            freeList = new HashSet();
            activeList = new HashSet();
    }
        
    // IMPLEMENTATION OF METHODS FROM THE Pool INTERFACE
        
    /**
     * @see org.objectweb.jonas.resource.pool.api.Pool#getTimeout
     */
    public long getTimeout() {
            return timeout;
    }
    
    /**
     * @see org.objectweb.jonas.resource.pool.api.Pool#getMinSize
     */
    public int getMinSize() {
            return minSize;
    }
    
    /**
     * @see org.objectweb.jonas.resource.pool.api.Pool#getMatchFactory
     */
    public PoolMatchFactory getMatchFactory() {
            return matchFactory;
    }
    
    /**
     * @see org.objectweb.jonas.resource.pool.api.Pool#getMaxSize
     */
    public int getMaxSize() {
            return maxSize;
    }
    
    /**
     * @see org.objectweb.jonas.resource.pool.api.Pool#getResource
     */
    public synchronized Object getResource(Object hints)
            throws Exception {
            if (matchFactory == null)
                    throw new Exception("The matchFactory mandatory!!");
            ManagedConnection res = null;
            if (!freeList.isEmpty()) {
                    try {
                            res = (ManagedConnection) matchFactory.matchResource(freeList, hints);
                    } 
                    catch (Exception ex) { }
            }
            if (res == null) {
                    if (curSize < maxSize) {
                            res = (ManagedConnection) matchFactory.createResource(hints);
                            activeList.add(res);
                            curSize++;
                    }
                    else if (freeSize > 0) {
                            Iterator it = freeList.iterator();
                            if (it.hasNext()) {
                                    res = (ManagedConnection) it.next();
                                    res.destroy();
                                    it.remove();
                                    curSize--;
                                    freeSize--;
                                    // Create a new one and return it
                                    res = (ManagedConnection) matchFactory.createResource(hints);
                                    activeList.add(res);
                                    curSize++;
                            }
                            else { 
                                    throw new Exception("No more resource available in pool!!");
                            }
                    }
                    else {
                            throw new Exception("No more resource available in pool!!");
                    }
            }
            else {
                    freeList.remove(res);
                    activeList.add(res);
                    freeSize--;
            }
//        printLists();
            return res;
    }
        
    public int getSize() {
            return curSize;
    }
        
    /**
     * @see org.objectweb.jonas.resource.pool.api.Pool#releaseResource
     */
    public synchronized void releaseResource(Object resource, boolean destroy)
        throws Exception {
            ManagedConnection res = (ManagedConnection) resource;
            if (matchFactory == null)
                    throw new Exception("The matchFactory mandatory!!");
            if (activeList == null)
                    throw new Exception("No active resources to releases!!");
            if (!activeList.contains(res))
                    throw new Exception("Attempt to release inactive resource!!");
            activeList.remove(res);
            if (curSize <= maxSize && !destroy) {
                    freeList.add(res);
                    freeSize++;
            }
            else {
                    res.destroy();
                    curSize--;
            }

            while (curSize < minSize) {
                    res = (ManagedConnection) matchFactory.createResource(null);
                    freeList.add(res);
                    curSize++;
                    freeSize++;
            }
//        printLists();
    }
        
    /**
     * @see org.objectweb.jonas.resource.pool.api.Pool#setTimeout
     */
    public synchronized void setTimeout(long crto) {
    }
    
    /**
     * @see org.objectweb.jonas.resource.pool.api.Pool#setMinSize
     */
    public synchronized void setMinSize(int minsize) throws Exception {
            if (matchFactory == null)
                    throw new Exception("The matchFactory mandatory!!");
            if ((minsize < 0) || (minsize > maxSize) || (minsize == minSize))
                    return;
            if (minsize < minSize) {
                minSize = minsize;
                    return;
            }
            ManagedConnection res;
            while (minsize > curSize) {
                    res = (ManagedConnection) matchFactory.createResource(null);
                    freeList.add(res);
                    curSize++;
                    freeSize++;
            }
            minSize = minsize;
    }
        
    /**
     * @see org.objectweb.jonas.resource.pool.api.Pool#setMinSize
     */
    public synchronized void setInitSize(int initsize) throws Exception {
            if (matchFactory == null)
                    throw new Exception("The matchFactory mandatory!!");
            if (initSize < 0 && initsize >= 0) {
                initSize = initsize;
                    setInitSize();
            }
    }

    /**
     * @see org.objectweb.jonas.resource.pool.api.Pool#setMinSize
     */
    private void setInitSize() throws Exception {
            if (initSize <= 0) return;
            int initsize = initSize < maxSize ? initSize : maxSize;
            initsize = initsize < minSize ? 0 : initsize - minSize;
            
            ManagedConnection res;
            for (int i=0;i<initsize;i++) {
                    res = (ManagedConnection) matchFactory.createResource(null);
                    freeList.add(res);
                    curSize++;
                    freeSize++;
            }
    }
    

    /**
     * @see org.objectweb.jonas.resource.pool.api.Pool#setMatchFactory
     */
    public synchronized void setMatchFactory(PoolMatchFactory pmf) {
            matchFactory = pmf;
    }
    
    /**
     * @see org.objectweb.jonas.resource.pool.api.Pool#setMaxSize
     */
    public synchronized void setMaxSize(int maxsize) throws Exception {
            if (matchFactory == null)
                    throw new Exception("The matchFactory mandatory!!");
            if ((maxsize < 0) || (maxsize < minSize) || (maxsize == maxSize))
                    return;
            if (maxsize < maxSize) {
                    int cursize = curSize;
                    int freesize = freeSize;
                    if (maxsize < cursize) {
                                // Removes as many free cells as possible
                                int nbRemove = cursize - maxsize;
                                if (freeList != null) {
                                        Iterator it = freeList.iterator();
                                        while (it.hasNext() && !freeList.isEmpty() && nbRemove > 0) {
                                                ManagedConnection res = (ManagedConnection) it.next();
                                                res.destroy();
                                                it.remove();
                                                nbRemove--;
                                                cursize--;
                                                freesize--;
                                        }
                                }
                    }
                    curSize = cursize;
                    freeSize = freesize;
            }
            maxSize = maxsize;
    }
        
//      void printLists() {
//              if (maxSize != 200)
//                      return;
//              System.err.println("minSize=" + minSize + ", maxSize=" + maxSize
//                                                 + ", curSize=" + curSize + ", freeSize=" +freeSize);
//              System.err.print("activeList:");
//              if (activeList == null)
//                      System.err.print(" null");
//              else {
//                      Iterator it = activeList.iterator();
//                      while (it.hasNext())
//                              System.err.print(" " + (ManagedConnection) it.next());
//              }
//              System.err.println(".");
//              System.err.print("freeList:");
//              if (freeList == null)
//                      System.err.print(" null");
//              else {
//                      Iterator it = freeList.iterator();
//                      while (it.hasNext())
//                              System.err.print(" " + (ManagedConnection) it.next());
//              }
//              System.err.println(".");
//      }

}
