/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Helene Joanin
 * Contributors: Philippe Durieux, Sebastien Chassande
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.genic;

import java.util.StringTokenizer;
import org.apache.velocity.app.VelocityEngine;
import org.objectweb.jonas_ejb.deployment.api.BeanDesc;
import org.objectweb.jonas_ejb.deployment.api.DeploymentCmp2Desc;
import org.objectweb.jonas_ejb.deployment.api.EntityCmp2Desc;
import org.objectweb.jonas_ejb.deployment.api.EntityDesc;
import org.objectweb.jonas_ejb.deployment.api.EntityJdbcCmp1Desc;
import org.objectweb.jonas_ejb.deployment.api.EntityJdbcCmp2Desc;
import org.objectweb.jonas_ejb.lib.BeanNaming;
import org.objectweb.jorm.comp.compiler.api.CompilerParameter;
import org.objectweb.jorm.comp.compiler.lib.BasicCompiler;
import org.objectweb.jorm.metainfo.api.Manager;
import org.objectweb.jorm.metainfo.api.Schema;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.wrapper.velocity.VelocityLogger;

import java.io.File;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

/**
 *  This class allows to generate the sources of:
 *  the class that implements the Enterprise bean's remote interface,
 *  the class that implements the Enterprise bean's home interface,
 *  the class that implements the Enterprise bean's local interface,
 *  the class that implements the Enterprise bean's localhome interface,
 *  the class of the Entity Handle in case of entity,
 *  and the extended class of the Bean for persistence in case of entity with CMP,
 * of a given Enterprise Java Bean.
 */
public class BeanSources {

    // Home interface. May be null if not defined. (for David)
    private String ejbHomeFileName = null;

    // Remote interface. May be null if not defined. (for David)
    private String ejbRemoteFileName = null;

    // EJBObject (Remote). May be null if no Remote interface.
    private String wrpRemoteFileName = null;

    // EJBLocalObject (Local). May be null if no Local interface.
    private String wrpLocalFileName = null;

    // Home. May be null if no Home interface.
    private String wrpHomeFileName = null;

    // LocalHome. May be null if no LocalHome interface.
    private String wrpLocalHomeFileName = null;

    // Handle (for Entity only)
    private String wrpHandleFileName = null;

    // Bean (for CMP Entity only)
    private String wrpBeanFileName = null;

    private String itfCohCmp2Entity = null;

    private String accessorBeanFileName = null;
    private String binderBeanFileName = null;
    private String pnameBeanFileName = null;
    private String pngBeanFileName = null;

    private String [] mapperNames = null;

    // Source objects used to generate classes
    private ArrayList sources;
    private ArrayList noRemoteJava;

    private String outdir = null; // directory where to place the generated files
    private BeanDesc dd = null;	// Bean Deployment Description
    private Manager manager = null;
    private Logger logger = null;

    /**
     * GenIC Constructor
     *
     * @param     beanDescr      deployment descriptor of the bean
     * @param     dirOutputName  path of the directory where to place the generated files
     *                           (empty string "" if the output directory is the current directory)
     * @exception GenICException In error case
     */
    public BeanSources(BeanDesc beanDesc, String dirOutputName, Manager manager, Logger log)
        throws GenICException {
        this(beanDesc, dirOutputName, log);
        this.manager = manager;
    }

    /**
     * GenIC Constructor
     *
     * @param     beanDescr      deployment descriptor of the bean
     * @param     dirOutputName  path of the directory where to place the generated files
     *                           (empty string "" if the output directory is the current directory)
     * @exception GenICException In error case
     */
    public BeanSources(BeanDesc beanDesc, String dirOutputName, Logger log) throws GenICException {

        logger = log;
        dd = beanDesc;
        outdir = dirOutputName;
        sources = new ArrayList();
        noRemoteJava = new ArrayList();

        // Create and initialize the Velocity Engine
        String jonasRoot = System.getProperty("install.root");
        if (jonasRoot == null) {
            throw new GenICException("System property 'install.root' not set");
        }
        String path2Tmpl = new String(jonasRoot + File.separatorChar + "templates"
            + File.separatorChar + "genic");

        VelocityEngine ve = new VelocityEngine();
        ve.setProperty(VelocityEngine.VM_LIBRARY, "");
        ve.setProperty(VelocityEngine.RESOURCE_LOADER, "file");
        ve.setProperty(VelocityEngine.FILE_RESOURCE_LOADER_PATH, path2Tmpl);
        if (logger != null) {
            // Connect the velocity log system to monolog
            ve.setProperty(VelocityEngine.RUNTIME_LOG_LOGSYSTEM, new VelocityLogger(logger));
        }
        try {
            ve.init();
        } catch (Exception e) {
            throw new GenICException("Cannot initialize the Velocity engine", e);
        }

        // generated Home
        if (dd.getHomeClass() != null) {
            ejbHomeFileName = dd.getHomeClass().getName().replace('.', File.separatorChar) + ".java";
            if (outdir.length() > 0) {
                ejbHomeFileName = outdir + File.separatorChar + ejbHomeFileName;
                wrpHomeFileName = outdir + File.separatorChar + BeanNaming.getPath(dd.getFullWrpHomeName());
            } else {
                wrpHomeFileName = dd.getWrpHomeName();
            }
            wrpHomeFileName = wrpHomeFileName.concat(".java");
            sources.add(new Source(dd, wrpHomeFileName, Source.HOME, ve, logger));
            // generated Handle for Remote Entity bean
            if (dd instanceof EntityDesc) {
                if (outdir.length() > 0) {
                    wrpHandleFileName = outdir + File.separatorChar + BeanNaming.getPath(dd.getFullWrpHandleName());
                } else {
                    wrpHandleFileName = dd.getWrpHandleName();
                }
                wrpHandleFileName = wrpHandleFileName.concat(".java");
                noRemoteJava.add(wrpHandleFileName);
                sources.add(new Source(dd, wrpHandleFileName, Source.ENTITY_HANDLE, ve, logger));
            }
        }

        // generated EJBObject (Remote)
        if (dd.getRemoteClass() != null) {
            ejbRemoteFileName = dd.getRemoteClass().getName().replace('.', File.separatorChar) + ".java";
            if (outdir.length() > 0) {
                ejbRemoteFileName = outdir + File.separatorChar + ejbRemoteFileName;
                wrpRemoteFileName = outdir + File.separatorChar + BeanNaming.getPath(dd.getFullWrpRemoteName());
            } else {
                wrpRemoteFileName = dd.getWrpRemoteName();
            }
            wrpRemoteFileName = wrpRemoteFileName.concat(".java");
            sources.add(new Source(dd, wrpRemoteFileName, Source.REMOTE, ve, logger));
        }

        // generated LocalHome
        if (dd.getLocalHomeClass() != null) {
            if (outdir.length() > 0) {
                wrpLocalHomeFileName = outdir + File.separatorChar + BeanNaming.getPath(dd.getFullWrpLocalHomeName());
            } else {
                wrpLocalHomeFileName = dd.getWrpLocalHomeName();
            }
            wrpLocalHomeFileName = wrpLocalHomeFileName.concat(".java");
            noRemoteJava.add(wrpLocalHomeFileName);
            sources.add(new Source(dd, wrpLocalHomeFileName, Source.LOCAL_HOME, ve, logger));
        }

        // generated EJBLocalObject (Local)
        if (dd.getLocalClass() != null) {
            if (outdir.length() > 0) {
                wrpLocalFileName = outdir + File.separatorChar + BeanNaming.getPath(dd.getFullWrpLocalName());
            } else {
                wrpLocalFileName = dd.getWrpLocalName();
            }
            wrpLocalFileName = wrpLocalFileName.concat(".java");
            noRemoteJava.add(wrpLocalFileName);
            sources.add(new Source(dd, wrpLocalFileName, Source.LOCAL, ve, logger));
        }

        // generated derived bean class for Entity Bean CMP
        if (dd instanceof EntityJdbcCmp1Desc) {
            if (outdir.length() > 0) {
                wrpBeanFileName = outdir + File.separatorChar + BeanNaming.getPath(dd.getFullDerivedBeanName());
            } else {
                wrpBeanFileName = dd.getDerivedBeanName();
            }
            wrpBeanFileName = wrpBeanFileName.concat(".java");
            noRemoteJava.add(wrpBeanFileName);
            sources.add(new Source(dd, wrpBeanFileName, Source.ENTITY_CMP_JDBC, ve, logger));
        }
        if (dd instanceof EntityCmp2Desc) {
            EntityCmp2Desc ecd = (EntityCmp2Desc) dd;
	    getAllMapperNames();
            if (outdir.length() > 0) {
                wrpBeanFileName = outdir + File.separatorChar
                    + BeanNaming.getPath(dd.getFullDerivedBeanName()) + ".java";
                accessorBeanFileName = outdir + File.separatorChar
                    + BeanNaming.getPath(ecd.getJormAccessorClassName()) + ".java";
		for (int i = 0; i < mapperNames.length; i++) {
		    String ma = mapperNames[i];
		    noRemoteJava.add(outdir + 
				     File.separatorChar +
				     BeanNaming.getPath(ecd.getJormMappingClassName(ma)) +
				     ".java");
		    noRemoteJava.add(outdir + 
				     File.separatorChar + 
				     BeanNaming.getPath(ecd.getJormBindingClassName(ma)) + 
				     ".java");
		}
                if (!ecd.hasPrimaryKeyField()) {
                    binderBeanFileName = outdir + File.separatorChar
                        + BeanNaming.getPath(ecd.getJormBinderClassName()) + ".java";
                    noRemoteJava.add(binderBeanFileName);
                    pnameBeanFileName = outdir + File.separatorChar
                        + BeanNaming.getPath(ecd.getJormPNameClassName()) + ".java";
                    noRemoteJava.add(pnameBeanFileName);
                    pngBeanFileName = outdir + File.separatorChar
                        + BeanNaming.getPath(ecd.getJormPNameGetterClassName()) + ".java";
                    noRemoteJava.add(pngBeanFileName);
                }
                if (ecd.needCoherenceHelper().booleanValue()) {
                    itfCohCmp2Entity = outdir + File.separatorChar
                        + BeanNaming.getPath(ecd.getCoherenceHelperFQItfName()) + ".java";
                }
            } else {
                accessorBeanFileName = BeanNaming.getBaseName(ecd.getJormAccessorClassName()) + ".java";
		for (int i = 0; i < mapperNames.length; i++) {
		    String ma = mapperNames[i];
		    noRemoteJava.add(BeanNaming.getBaseName(ecd.getJormMappingClassName(ma)) + ".java");
		    noRemoteJava.add(BeanNaming.getBaseName(ecd.getJormBindingClassName(ma)) + ".java");
		}
                wrpBeanFileName = BeanNaming.getPath(dd.getDerivedBeanName()) + ".java";
                if (!ecd.hasPrimaryKeyField()) {
                    pnameBeanFileName =
                        BeanNaming.getBaseName(ecd.getJormPNameClassName()) + ".java";
                    noRemoteJava.add(pnameBeanFileName);
                    pngBeanFileName =
                        BeanNaming.getBaseName(ecd.getJormPNameGetterClassName()) + ".java";
                    noRemoteJava.add(pngBeanFileName);
                    binderBeanFileName =
                        BeanNaming.getBaseName(ecd.getJormBinderClassName()) + ".java";
                    noRemoteJava.add(binderBeanFileName);
                }
                if (ecd.needCoherenceHelper().booleanValue()) {
                    itfCohCmp2Entity =
                        BeanNaming.getPath(ecd.getCoherenceHelperItfName()) + ".java";
                }
            }
            if (ecd.needCoherenceHelper().booleanValue()) {
                sources.add(new Source(
                    dd, itfCohCmp2Entity, Source.ITF_COH_CMP2_ENTITY, ve, logger));
                noRemoteJava.add(itfCohCmp2Entity);
            }
            noRemoteJava.add(accessorBeanFileName);
            noRemoteJava.add(wrpBeanFileName);
            sources.add(new Source(dd, wrpBeanFileName, Source.ENTITY_CMP2_JDBC, ve, logger));
        }
    }

    /**
     * TODO: do not use a system property for this ???
     */
    private void getAllMapperNames() {
	String mns = System.getProperty("JONAS_CMP_MAPPERNAMES");
	if (mns != null && mns.length() > 0) {
	    StringTokenizer st = new StringTokenizer(mns, ",", false);
	    mapperNames = new String[st.countTokens()];
	    int idx = 0;
	    while (st.hasMoreElements()) {
		String mapper = st.nextToken().trim();
		mapperNames[idx++] = mapper;
	    }
	} else {
	    mapperNames = new String[]{"rdb"};
	}
    }

    /**
     * Generates the java sources
     */
    public void generate() throws GenICException {

        // Generates the java sources
        for (Iterator it = sources.iterator(); it.hasNext();) {
            Source src = (Source) it.next();
            src.generate();
        }

	// In case of Cmp2, call jorm compiler for each mapper.
        if (dd instanceof EntityCmp2Desc) {
            EntityCmp2Desc ecd = (EntityCmp2Desc) dd;
            String cn = ecd.getJormName();
            Schema schema = manager.getSchema(ecd.getJormSchemaName());
            ArrayList mos = new ArrayList(2);
            BasicCompiler jc = new BasicCompiler();
            CompilerParameter cp = jc.getCompilerParameter();
            cp.setLogger(logger);
            cp.setMappingName(DeploymentCmp2Desc.MAPPING_NAME);
            cp.setNoCompile(false);
            cp.setKeepSrc(true);
            cp.setOutputDirectory(outdir);
            cp.setPBindingInheritance("org.objectweb.jonas_ejb.container.JEntitySwitch");
            mos.add(schema.getClass(getClassName(cn)));
            if (!ecd.hasPrimaryKeyField()) {
                mos.add(schema.getCompositeName(getClassName(ecd.getJormCompositeName())));
            }
	    for (int i = 0; i < mapperNames.length; i++) {
		String mappername = mapperNames[i];
		logger.log(BasicLevel.DEBUG, "Calling Jorm Compiler with Mapper "+mappername);
		cp.setMapperName(mappername);
		cp.setPClassMappingInheritance(getFactoryName(mappername));
		try {
		    jc.generateFiles(mos);
		} catch (Exception e) {
		    throw new GenICException("Problem during jorm generation", e);
		}
	    }
	}
    }

    /**
     * Return the bean's name
     */
    public String getEjbName() {
        return dd.getEjbName();
    }

    /**
     * Return the file name of the generated source for the Home
     * (null if none)
     */
    public String getWrpHomeFileName() {
        return wrpHomeFileName;
    }

    /**
     * Return the file name of the generated source for the Remote
     * (null if none)
     */
    public String getWrpRemoteFileName() {
        return wrpRemoteFileName;
    }

    /**
     * Return the 'pseudo' source file name of the Remote interface
     * (DAVID only used: The Stub file name depends on this name)
     */
    public String getEjbRemoteFileName() {
        return ejbRemoteFileName;
    }

    /**
     * Return the 'pseudo' source file name of the Home interface
     * (DAVID only used: The Stub file name depends on this name)
     */
    public String getEjbHomeFileName() {
        return ejbHomeFileName;
    }

    /**
     * Return the class name of the generated source for the Remote (package included)
     */
    public String getWrpRemoteClassName() {
        return dd.getFullWrpRemoteName();
    }

    /**
     * Return the class name of the generated source for the Home (package included)
     */
    public String getWrpHomeClassName() {
        return dd.getFullWrpHomeName();
    }

    public Collection getNoRemoteJavas() {
        return noRemoteJava;
    }

    public String getPackageName(String fqcn) {
        if (fqcn == null)
            return null;
        int idx = fqcn.lastIndexOf('.');
        return (idx == -1 ? "" : fqcn.substring(0, idx));
    }

    public String getClassName(String fqcn) {
        if (fqcn == null)
            return null;
        int idx = fqcn.lastIndexOf('.');
        int len = fqcn.length();
        return (idx == -1
            ? fqcn
            : (idx != (len - 1)
            ? fqcn.substring(idx + 1, fqcn.length())
            : ""));
    }

    public String getFactoryName(String mapperName) {
        String s = mapperName;
        int idx = s.indexOf(".");
        if (idx != -1)
            s = s.substring(0, idx);
        return "org.objectweb.jonas_ejb.container.jorm."
            + Character.toUpperCase(s.charAt(0)) + s.substring(1, s.length())
            + "Factory";
    }
}
